# Version 0.1
#
#  BookAnnotation is copyright 2005 Jeffrey G. Pittman.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#  MA 02111-1307 USA

from Products.Archetypes.public import BaseFolderSchema, Schema
from Products.Archetypes.public import StringField, IntegerField, FloatField, LinesField, BooleanField
from Products.Archetypes.public import LinesWidget, TextAreaWidget, IdWidget, IntegerWidget, DecimalWidget, StringWidget, BooleanWidget, MultiSelectionWidget
from Products.Archetypes.public import BaseFolder, registerType
from Products.BookAnnotation.config import PROJECTNAME
from Products.CMFCore import CMFCorePermissions

import sys,string,cStringIO

from config import *
from AccessControl import ClassSecurityInfo, ModuleSecurityInfo
from reportlab.pdfgen.canvas import Canvas
from Products.BookAnnotation.sample_data import sample_book_annotations


def colorstr(rgb):
    """
    from http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/325823
    """
    return "#%x%x%x" % (rgb[0]/16,rgb[1]/16,rgb[2]/16)
    
class Annotation:
    def __init__(self, chapter, page, line, left_position, term_or_phrase, text, red, green, blue):
        self.chapter = chapter
        self.page = page
        self.line = line
        self.left_position = left_position

        # since dashes are used for background, and are keyed on in
        # processing, dashes in term_or_phrase, if they occur, should 
        # be replaced with underscores 
        self.term_or_phrase = term_or_phrase.replace('-','_')
        self.text = text
        self.color = (red, green, blue)
        self.neighbors = []
        self.line_text = ""

    def hasNeighbors(self):
        if len(self.neighbors) > 0:
            return 1
        else:
            return 0

    def addNeighbor(self, neighbor):
        if self != neighbor and neighbor not in self.neighbors:
            self.neighbors.append(neighbor)
        # should raise an exception if attempt to set self as neighbor, i suppose?

    def setLineText(self, line_text):
        self.line_text = line_text

    def isLineTextSet(self):
        if len(self.line_text) > 0:
            return 1
        else:
            return 0

class BookAnnotationRenderingEngine:
    def __init__(self,
                 line_length,
                 background_character,
                 lines_per_page,
                 annotation_input):
        self.line_length = line_length
        self.background_character = background_character
        self.lines_per_page = lines_per_page
        self.annotation_input = annotation_input
        
        # ordered lists -- the order of the CSV -- user determines order,
        # by the order of input in the CSV data
        self.chapters_in_order = []
        self.pages_in_order = []

        # dictionaries -- makes work and lookup easier
        self.chapters_and_pages = {}
        self.pages_and_annotations = {}

        # create annotation objects from CSV data and fill lists and
        # dictionaries.
        for cvs_input_line in self.annotation_input:
            items = eval('[%s]' % cvs_input_line) # we've validated by now, so good to go
            chapter, page, line, left_position, term_or_phrase, annotation, red, green, blue = items

            if chapter not in self.chapters_in_order:
                self.chapters_in_order.append(chapter)

            if page not in self.pages_in_order:
                self.pages_in_order.append(page)

            if not self.chapters_and_pages.has_key(chapter):
                self.chapters_and_pages[chapter] = []
                self.chapters_and_pages[chapter].append(page)
            else:
                if page not in self.chapters_and_pages[chapter]:
                    self.chapters_and_pages[chapter].append(page)

            # input allows left_position to go from 0 to self.line_length, 
            # but here allow up to self.line_length - 1, for indexing math
            if left_position == self.line_length: 
                left_position = self.line_length - 1

            annotation = Annotation(chapter, 
                                    page, 
                                    line, 
                                    left_position, 
                                    term_or_phrase, 
                                    annotation, 
                                    red, green, blue)

            # set up some handy dictionaries
            if not self.pages_and_annotations.has_key(page):
                self.pages_and_annotations[page] = []
                self.pages_and_annotations[page].append(annotation)
            else:
                self.pages_and_annotations[page].append(annotation)

        self.chapter_count = len(self.chapters_and_pages.keys())

        self.page_count = len(self.pages_and_annotations.keys())

        #for chapter in self.chapters_and_pages.keys():
            #print 'chapter %d' % chapter, 'pages: %s' % str(self.chapters_and_pages[chapter])

        #print ' '

        #for page in self.pages_and_annotations.keys():
            #print 'page %d' % page, 'annotations on lines: %s' % str([('%d ' % a.line) for a in self.pages_and_annotations[page]])

        # set neighbors on the same lines
        current_line = -1
        current_neighbors = []
        for key in self.pages_and_annotations.keys():
            for annotation in self.pages_and_annotations[key]:
                line = annotation.line
                if line == current_line or current_line == -1:
                    current_neighbors.append(annotation)
                else:
                    for a in current_neighbors:
                        for neighbor in current_neighbors:
                            if a != neighbor:
                                neighbor.addNeighbor(a)
                                a.addNeighbor(neighbor)
                    current_neighbors = []
                    current_neighbors.append(annotation)
                current_line = line
            for annotation in current_neighbors:
                for neighbor in current_neighbors:
                    if annotation != neighbor:
                        neighbor.addNeighbor(annotation)
                        annotation.addNeighbor(neighbor)

    def getChapters(self):
        return self.chapters_in_order
        
    def getPages(self):
        return self.pages_in_order
        
    def getHTML(self):
        var = []
        for chapter in self.chapters_in_order:
            var += self.getHTMLForChapter(chapter)

        return var

    def getHTMLForChapter(self, chapter):
        var = []
        var += '<H1>%s</H1>\n' % chapter
        for page in self.chapters_and_pages[chapter]:
            var += self.getHTMLForPage(page)

        return var

    def getHTMLForListing(self):
        var = []
        for chapter in self.chapters_in_order:
            var += '<H1>%s</H1>\n' % chapter
            for page in self.chapters_and_pages[chapter]:
                var += '<div style=\"text-align:right;\">Page %s</div>\n' % page
                var += '<dl>\n'
                for annotation in self.pages_and_annotations[page]:
                    var += '<dt>%d: %s</dt>\n' % (annotation.line, annotation.term_or_phrase)
                    var += '<dd>%s</dd>\n' % annotation.text
                var += '</dl>\n'

        return var

    def getHTMLForPage(self, page):
        #var = ["<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n",
               #"<html>\n",
               #"<head>\n",
               #"  <title>Book Annotation, Basic Listing</title>\n",
               #"</head>\n",
               #"<body>\n",
        var = ["<div style=\"font-family: courier;background-color:#FFFFFF;\">\n"]

        line_text = ''.join([self.background_character * self.line_length, '<br />\n'])

        page_text = []
        for i in range(self.lines_per_page):
            page_text.append(line_text)

        var += '<br />'
        var += '<div style = \"text-align:right;\">Page %s</div>' % page
        for annotation in self.pages_and_annotations[page]:
            if not annotation.isLineTextSet():
                if annotation.hasNeighbors():
                    # the object is to fit the neighbors within the
                    # self.line_length count of characters,
                    # along with the required tags around them.

                    # first, in the case of an absolute ill-fitting
                    # case, where the sum total length of the
                    # neighboring term_or_phrases exceeds self.line_length, break out
                    # with a simple oversized line (this worst case
                    # could happen if one or more term_or_phrases were
                    # entered as longer text than actually present in
                    # the book).
                    items_to_do = []
                    items_to_do.append(annotation)
                    for a in annotation.neighbors:
                        items_to_do.append(a)
                    sum_total_length = 0
                    for a in items_to_do:
                        sum_total_length += len(a.term_or_phrase)

                    # order the annotations in terms of
                    # left_position
                    items_ordered = []
                    while len(items_ordered) < len(items_to_do):
                        minimum_left_position = self.line_length
                        a_with_minimum = None
                        for j in items_to_do:
                            if j not in items_ordered:
                                if j.left_position < minimum_left_position:
                                    a_with_minimum = j
                                    minimum_left_position = j.left_position
                        items_ordered.append(a_with_minimum)

                    # term_or_phrases could also overlap upon one 
                    # another, so check for that also, and punt, also
                    overlapping_condition = 0
                    previous_end = -1
                    for a in items_ordered:
                        if (self.line_length - a.left_position) < len(a.term_or_phrase):
                            a.left_position = self.line_length - len(a.term_or_phrase)
                        end_position = (a.left_position-1) + len(a.term_or_phrase)
                        if previous_end != -1:
                            if previous_end >= a.left_position:
                                overlapping_condition = 1
                        previous_end = end_position

                    if sum_total_length > self.line_length or overlapping_condition:
                        # build the final simple oversized line text
                        simple_oversized_line = []
                        for a in items_ordered:
                            simple_oversized_line.append(self.background_character)
                            simple_oversized_line.append('<a href=\"#\" onmouseover=\"return escape(\'')
                            simple_oversized_line.append(a.text)
                            simple_oversized_line.append('\')\">')
                            simple_oversized_line.append(a.term_or_phrase)
                            simple_oversized_line.append('</a>')
                            simple_oversized_line.append(self.background_character)

                        # set the final line text
                        page_text[a.line-1] = ''.join([simple_oversized_line,'<br />\n'])

                        # set the final line text in the annotations
                        for a in items_ordered:
                            a.setLineText(page_text[a.line-1])
                        break;

                    # if we make it here, we have space for all

                    # reuse working_line_text so term_or_phrases are
                    # all plotted within dashes of a single line.
                    working_line_text = line_text
                    for a in items_ordered:
                        if (self.line_length - a.left_position) < len(a.term_or_phrase):
                            a.left_position = self.line_length - len(a.term_or_phrase)
                        end_position = (a.left_position-1) + len(a.term_or_phrase)
                        working_line_text = working_line_text[0:a.left_position] + \
                                            a.term_or_phrase + \
                                            working_line_text[end_position:(self.line_length - 1)]

                    # now go through and insert the tags at the
                    # term_or_phrase boundaries; there should be n * 2
                    # boundaries to find, unless there is overlap of
                    # term_or_phrases or unless a term_or_phrase starts at
                    # 0 or ends at self.line_length - 1.
                    position = 0
                    final_line_text = []
                    positions_after_term_or_phrase = 0
                    for a in items_ordered:
                        while position < (self.line_length - 1) and working_line_text[position] == self.background_character:
                            final_line_text.append(self.background_character)
                            position += 1
                        final_line_text.append('<a href=\"#\" onmouseover=\"return escape(\'')
                        final_line_text.append(a.text)
                        final_line_text.append('\')\">')
                        final_line_text.append(a.term_or_phrase)
                        final_line_text.append('</a>')

                        # advance position to next term_or_phrase;
                        # positions_after_term_or_phrase will tell us the
                        # number of dashes that should go after the
                        # last term_or_phrase
                        while position < (self.line_length - 1) and working_line_text[position] != self.background_character:
                            position += 1
                        # we are at the beginning of the next dashes
                        # section, so advance beyond to calculate
                        # positions_after_term_or_phrase
                        positions_after_term_or_phrase = position
                        position2 = position
                        while position2 < (self.line_length - 1) and working_line_text[position2] == self.background_character:
                            position2 += 1
                        positions_after_term_or_phrase = position2 - positions_after_term_or_phrase
                    
                    # add dashes after last term_or_phrase
                    if positions_after_term_or_phrase > 0:
                        final_line_text.append(self.background_character * positions_after_term_or_phrase)

                    # add the terminating tag
                    final_line_text.append('<br />\n')
                    
                    # set the final result of all that business
                    page_text[annotation.line-1] = ''.join(final_line_text)
                else:
                    if (self.line_length - annotation.left_position) < len(annotation.term_or_phrase):
                        annotation.left_position = self.line_length - len(annotation.term_or_phrase)
                    end_position = (annotation.left_position-1) + len(annotation.term_or_phrase)
                    page_text[annotation.line-1] = (self.background_character * annotation.left_position) + \
                                                   '<a href=\"#\" onmouseover=\"return escape(\'' + \
                                                   annotation.text + \
                                                   '\')\">' + \
                                                   annotation.term_or_phrase + \
                                                   '</a>' + \
                                                   (self.background_character * (self.line_length - end_position - 1)) + \
                                                   '<br />\n'
        # add the finished lines
        for i in range(len(page_text)):
            var += "%2d: %s" % (i+1, page_text[i])

        var += "</div>\n",
        #var += "</body>\n",
        #var += "</html>\n",

        return var

class BookAnnotationProcessor:
    def __init__(self, 
                 render_types,
                 line_length,
                 background_character,
                 lines_per_page,
                 book_annotations_input):
        self.render_types = render_types
        self.line_length = line_length
        self.background_character = background_character
        self.lines_per_page = lines_per_page
        self.book_annotations_input = book_annotations_input

        self.book_annotation = BookAnnotationRenderingEngine(self.line_length,
                                                             self.background_character, 
                                                             self.lines_per_page, 
                                                             self.book_annotations_input)

    def getHTML(self):
        return self.book_annotation.getHTML()

    def getHTMLForChapter(self, chapter):
        return self.book_annotation.getHTMLForChapter(chapter)

    def getHTMLForPage(self, page):
        return self.book_annotation.getHTMLForPage(page)

    def getHTMLForListing(self):
        return self.book_annotation.getHTMLForListing()

    def getChapters(self):
        return self.book_annotation.getChapters()

    def getPages(self):
        return self.book_annotation.getPages()

book_annotation_schema = BaseFolderSchema +  Schema((
     StringField('id',
                 required=0, ## Still actually required, but
                             ## the widget will supply the missing value
                             ## on non-submits
                 mode="rw",
                 accessor="getId",
                 mutator="setId",
                 default=None,
                 widget=IdWidget(label="Short Name",
                                 description="Should not contain spaces, underscores or mixed case. "\
                                 "Short Name is part of the item's web address.",
                                 visible={'view' : 'visible'}),
                 ),
    StringField('title',
                required=1,
                searchable=1,
                default='',
                accessor='Title',
                widget=StringWidget(label="Title",
                                    description="This name shows in the navigation menu."),
                ),
    StringField('description',
                isMetadata=1,
                accessor='Description',
                widget=TextAreaWidget(label='Description', 
                              description='This description is optional, but not a bad idea for search text purposes.'),
                ),
    StringField('render_types',
                vocabulary=RENDER_TYPES,
                mutator="setRenderTypes",
                accessor="RenderTypes",
                widget=MultiSelectionWidget(label='Render Types',
                                            description='Select render types to generate.',
                                            format='checkbox'),
                ),
    IntegerField('line_length',
                 required=1,
                 default='65',
                 validators=("isInt"),
                 accessor='LineLength',
                 widget=IntegerWidget(label="Line Length",
                                      description="Count characters on a sample of lines and use the maximum."),
                 ),
    StringField('background_character',
                required=1,
                default='-',
                accessor="BackgroundCharacter",
                widget=StringWidget(label='Background Character',
                                    maxlength=1,
                                    size=1,
                                    description='Character to use for showing the lines of the page.'),
                ),
    IntegerField('lines_per_page',
                 required=1,
                 default='35',
                 validators=("isInt"),
                 accessor='LinesPerPage',
                 widget=IntegerWidget(label="Lines Per Page",
                                      description="Lines per page."),
                 ),
    LinesField('book_annotations_field',
               required=0,
               default='\n'.join(sample_book_annotations),
               accessor="BookAnnotations",
               validators=('validate_book_annotations',),
               widget=LinesWidget(label='Book Annotations',
                                  description='For each book_annotation, provide: chapter, page, line, % from left, term/phrase, annotation, red, green, blue',),
               ),
        ))


class BookAnnotation(BaseFolder):
    """
    A folderish object to hold book annotations.
    """
    allowed_content_types = ('Document','Image')

    filter_content_types = 1

    global_allow = 1

    schema = book_annotation_schema

    content_icon = 'bookannotationicon.gif'

    security = ClassSecurityInfo()

    actions = ({'id': 'view',
                'name': 'View',
                'action': 'string:${object_url}/BookAnnotation_view',
                'permissions': (CMFCorePermissions.View,)},
               {'id': 'references',
                'name': 'References',
                'action': 'string:${object_url}/reference_edit',
                'permissions': (CMFCorePermissions.ModifyPortalContent,),
                'visible':0},
               {'id': 'metadata',
                'name': 'Properties',
                'action': 'string:${object_url}/base_metadata',
                'permissions': (CMFCorePermissions.ModifyPortalContent,),
                'visible':0})

    def gatherInput(self):
        render_types = self.RenderTypes()
        line_length = self.LineLength()
        background_character = self.BackgroundCharacter()
        lines_per_page = self.LinesPerPage()
        book_annotations_input = self.BookAnnotations()

        return (render_types, 
                line_length,
                background_character,
                lines_per_page,
                book_annotations_input)
            
    # processForm is called for creation
    def processForm(self, data=1, metadata=0, REQUEST=None, values=None):
        ret = BaseFolder.processForm(self, data, metadata, REQUEST, values)

        self.createBookAnnotationProcessor(self.gatherInput())

        return ret

    # update is called for edit
    def update(self, **kwargs):
        ret = BaseFolder.update(self, **kwargs)

        self.createBookAnnotationProcessor(self.gatherInput())

        return ret

    def createBookAnnotationProcessor(self, input_items):
        render_types_with_messages, line_length, background_character, lines_per_page, book_annotations_input = input_items

        # trim off the parenthetic messages in the render_types
        render_types = []
        for render_type in render_types_with_messages:
            if len(render_type) > 0: # extra one returned
                render_types.append(render_type.split()[0])

        self.book_annotation_processor = BookAnnotationProcessor(render_types, 
                                                                 line_length,
                                                                 background_character,
                                                                 lines_per_page,
                                                                 book_annotations_input)
    
        if 'COMPLETE' in render_types:
            existing_file = getattr(self, 'complete.html', None)
            if existing_file:
                 self.manage_delObjects('complete.html')

            self.invokeFactory('Document', id='complete.html')
            html_file = self['complete.html']
            html_file.edit(text_format='html',text=''.join(self.book_annotation_processor.getHTML()))

        if 'CHAPTERS' in render_types:
            for chapter in self.getChapters():
                chapter_filename = self.getChapterFilename(chapter)
                existing_file = getattr(self, chapter_filename, None)
                if existing_file:
                    self.manage_delObjects(chapter_filename)
                    
                self.invokeFactory('Document', id=chapter_filename)
                html_file = self[chapter_filename]
                html_file.edit(text_format='html',text=''.join(self.book_annotation_processor.getHTMLForChapter(chapter)))

        if 'PAGES' in render_types:
            for page in self.getPages():
                page_filename = self.getPageFilename(page)
                existing_file = getattr(self, page_filename, None)
                if existing_file:
                    self.manage_delObjects(page_filename)
                    
                self.invokeFactory('Document', id=page_filename)
                html_file = self[page_filename]
                html_file.edit(text_format='html',text=''.join(self.book_annotation_processor.getHTMLForPage(page)))

        if 'LISTING' in render_types:
            existing_file = getattr(self, 'listing.html', None)
            if existing_file:
                 self.manage_delObjects('listing.html')

            self.invokeFactory('Document', id='listing.html')
            html_file = self['listing.html']
            html_file.edit(text_format='html',text=''.join(self.book_annotation_processor.getHTMLForListing()))

    def getChapters(self):
        return self.book_annotation_processor.getChapters()

    def getChapterFilenames(self):
        chapter_filenames = []
        for chapter in self.book_annotation_processor.getChapters():
            chapter_filenames.append(self.getChapterFilename(chapter))
        
        return chapter_filenames

    def getPages(self):
        return self.book_annotation_processor.getPages()

    def getPageFilenames(self):
        page_filenames = []
        for page in self.book_annotation_processor.getPages():
            page_filenames.append(self.getPageFilename(page))
        
        return page_filenames

    def getFirstChapterFilename(self):
        first_chapter = self.getChapters()[0]
        return self.getChapterFilename(first_chapter)

    def getChapterFilename(self, chapter):
        chapter_filename = '%s.html' % chapter
        chapter_filename = chapter_filename.replace(' ', '')
        return chapter_filename

    def getFirstPageFilename(self):
        first_page = self.getPages()[0]
        return self.getPageFilename(first_page)

    def getPageFilename(self, page):
        page_filename = '%s.html' % page
        page_filename = page_filename.replace(' ', '')
        return page_filename

registerType(BookAnnotation, PROJECTNAME)


