from Products.validation.interfaces import ivalidator

class BookAnnotationsValidator:

    __implements__ = (ivalidator,)

    def __init__(self, name):
        self.name = name

# chapter, page, line, term_or_phrase, annotation, red, green, blue
    def __call__(self, value, *args, **kwargs):
        self.lines = {'chapter'        : [],
                      'page'           : [],
                      'line'           : [],
                      'left_position'  : [],
                      'term_or_phrase' : [],
                      'annotation'     : [],
                      'red'            : [],
                      'green'          : [],
                      'blue'           : []}
                  
        # setup and check for item count errors:

        line_errors = []
        line_count = 1
        for book_annotation_string in value:
            chapter = ""
            page = ""
            line = ""
            left_position = ""
            term_or_phrase = ""
            annotation= ""
            red = ""
            green = ""
            blue = ""
            try:
                items = eval('[%s]' % book_annotation_string)
            except:
                return "Mixed numeric and text input somewhere on line %d" % line_count
            if len(items) != 9:
                line_errors.append("%d" % line_count)
            else:
                chapter, page, line, left_position, term_or_phrase, annotation, red, green, blue = items 
                self.lines['chapter'].append(chapter)
                self.lines['page'].append(page)
                self.lines['line'].append(line)
                self.lines['left_position'].append(left_position)
                self.lines['term_or_phrase'].append(term_or_phrase)
                self.lines['annotation'].append(annotation)
                self.lines['red'].append(red)
                self.lines['green'].append(green)
                self.lines['blue'].append(blue)
            line_count += 1
        if len(line_errors) > 0:
            return "Wrong number of items on lines: %s" % line_errors

        # check for basic conversion errors:

        error_message = ""
        errors = None

        errors = self.checkForChapterErrors()
        if errors:
            error_message = "chapter (field 1) error on lines: %s" % str(errors)
            return error_message

        errors = self.checkForPageErrors()
        if errors:
            error_message = "page (field 2) error on lines: %s" % str(errors)
            return error_message

        errors = self.checkForLineErrors()
        if errors:
            error_message = "line (field 3) error on lines: %s" % str(errors)
            return error_message

        errors = self.checkForLeftPositionErrors()
        if errors:
            error_message = "left_position (field 4) error on lines: %s" % str(errors)
            return error_message

        errors = self.checkForTermOrPhraseErrors()
        if errors:
            error_message = "term_or_phrase (field 5) error on lines: %s" % str(errors)
            return error_message

        errors = self.checkForAnnotationErrors()
        if errors:
            error_message = "annotation (field 6) error on lines: %s" % str(errors)
            return error_message

        errors = self.checkForRedErrors()
        if errors:
            error_message = "Red (field 9) error on lines: %s" % str(errors)
            return error_message

        errors = self.checkForGreenErrors()
        if errors:
            error_message = "Green (field 10) error on lines: %s" % str(errors)
            return error_message

        errors = self.checkForBlueErrors()
        if errors:
            error_message = "Blue (field 11) error on lines: %s" % str(errors)
            return error_message

        # check for color range errors:
        errors = self.checkForColorRangeErrors()
        if errors:
            error_message = "One or more color values out of range (0-255) on lines: %s" % str(errors)
            return error_message

        return 1

    def checkForChapterErrors(self):
        return None

    def checkForPageErrors(self):
        return None

    def checkForLineErrors(self):
        line_errors = []
        count = 1
        for line in self.lines['line']:
            try:
                int(line)
            except:
                line_errors.append(count)
            if int(line) < 0:
                line_errors.append(count)
            count += 1
        if len(line_errors) > 0:
            return line_errors
        return None

    def checkForLeftPositionErrors(self):
        line_errors = []
        count = 1
        for left_position in self.lines['left_position']:
            try:
                int(left_position)
            except:
                line_errors.append(count)
            if int(left_position) < 0 or int(left_position) > 100:
                line_errors.append(count)
            count += 1
        if len(line_errors) > 0:
            return line_errors
        return None

    def checkForTermOrPhraseErrors(self):
        # pass
        return None

    def checkForAnnotationErrors(self):
        # pass
        return None

    def checkForColorErrors(self, color_name):
        line_errors = []
        count = 1
        for color in self.lines[color_name]:
            try:
                int(color)
            except:
                line_errors.append(count)
        if len(line_errors) > 0:
            return line_errors
        return None

    def checkForRedErrors(self):
        return self.checkForColorErrors('red')

    def checkForGreenErrors(self):
        return self.checkForColorErrors('green')

    def checkForBlueErrors(self):
        return self.checkForColorErrors('blue')

    def checkForColorRangeErrors(self):
        line_errors = []
        for i in range(len(self.lines['red'])):
            red = int(self.lines['red'][i])
            green = int(self.lines['green'][i])
            blue = int(self.lines['blue'][i])

            # must be between 0 and 255
            if red < 0 or red > 255 or green < 0 or green > 255 or blue < 0 or blue > 255:
                line_errors.append(i+1)
        if len(line_errors) > 0:
            return line_errors
        return None

